import copy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import rcParams
from typing import Optional, Union, Any, Mapping, Literal, Tuple
from types import MappingProxyType
from matplotlib.gridspec import GridSpec
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.colors import LinearSegmentedColormap


def get_s(pvalue, s) -> Tuple[int, str]:
    """
    Adjusts scatter point size and type based on p-value.
    """
    if pvalue > 0.5:
        new_s = s
        type = '> 0.5'
    elif 0.5 > pvalue > 0.1:
        new_s = s*2
        type = '> 0.1'
    elif 0.1 > pvalue > 0.05:
        new_s = s * 3
        type = '> 0.05'
    else:
        new_s = s * 4
        type = '< 0.05'
    return new_s, type


def get_c(score, vmin, vmax):
    """
    Adjusts the score based on given thresholds.

    Parameters
    ----------
        score : float
            Score.
        vmin : float, optional
            Minimum threshold value.
        vmin : float, optional
            Maximum threshold value.

    Returns
    -------
        float: Adjusted score.
    """
    if vmax is not None and score > vmax:
        score = vmax
    if vmin is not None and score < vmin:
        score = vmin
    return score


def _dotplot(
        scores: pd.DataFrame,
        pvalue: pd.DataFrame,
        s_cutoff: float = 0.05,
        p_cutoff: float = 3.0,
        figsize = (6.4, 5.5),
        vmax: float = None,
        vmin: float = None,
        ax=None,
        cmap=None,
        sig_color='#000000',
        s: int = 10,
        label_fontsize=10,
        **kwargs
) -> Optional[Axes]:
    """
    Main underlying helper function for generating heatmaps.

    Parameters
    ----------
    scores : pd.DataFrame
        Score DataFrame.
    pvalue : pd.DataFrame
        P-value DataFrame.
    s_cutoff : float, optional
        Cutoff value for scatter point size.
    p_cutoff : float, optional
        Cutoff value for p-values.
    figsize : Tuple[float, float], optional
        Figure size.
    vmax : float, optional
        Maximum threshold value.
    vmin : float, optional
        Minimum threshold value.
    ax : Axes, optional
        Matplotlib Axes object.
    cmap : str, optional
        Color map.
    sig_color : str, optional
        Color for significant points.
    s : int, optional
        Initial scatter point size.
    label_fontsize : int, optional
        Font size for labels.
    kwargs: Any, optional
        Other params of ax.scatter()

    Returns
    -------
    Axes
        Generated Axes object.
    """
    if type(cmap) == type(None):
        cmap = "Spectral_r"

    if type(ax) == type(None):
        fig, ax = plt.subplots(figsize=figsize)

    n_rows = scores.shape[0] * scores.shape[1]
    flat_df = pd.DataFrame(index=list(range(n_rows)), columns=['x', 'y', 'value', 'p', 'type', 'sig'])
    i = 0
    for index_ct, ct in enumerate(scores.columns):
        for index_lr, lr in enumerate(scores.index):
            new_p, label = get_s(pvalue.values[index_lr, index_ct], s)
            score = get_c(scores.values[index_lr, index_ct], vmin=vmin, vmax=vmax)
            if pvalue.values[index_lr, index_ct] < p_cutoff and scores.values[index_lr, index_ct] > s_cutoff:
                flat_df.iloc[i, :] = [ct, lr, score, new_p, label, sig_color]
            else:
                flat_df.iloc[i, :] = [ct, lr, score, new_p, label, '#FFFFFF']
            i += 1

    x = flat_df['x']
    y = flat_df['y']

    x_labels = list(x.values)
    y_labels = list(y.values)
    x_to_num = {p[1]: p[0] for p in enumerate(x_labels)}
    y_to_num = {p[1]: p[0] for p in enumerate(y_labels)}

    out = ax.scatter(
        x=x.map(x_to_num),  # Use mapping for x
        y=y.map(y_to_num),  # Use mapping for y
        c=flat_df['value'].tolist(),
        s=flat_df['p'].tolist(),
        edgecolor=flat_df['sig'].tolist(),
        cmap=cmap,
        marker="o",
        **kwargs
    )

    out.set_array(flat_df['value'].values.astype(int))
    out.set_clim(min(flat_df['value']), max(flat_df['value']))
    cbar = plt.colorbar(out)
    cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.set_ylabel("lr score", rotation=270)

    # Show column labels on the axes
    ax.set_xticks([x_to_num[v] for v in np.unique(x_labels)])
    ax.set_xticklabels(np.unique(x_labels), rotation=45, horizontalalignment="right", fontsize=label_fontsize)
    ax.set_yticks([y_to_num[v] for v in np.unique(y_labels)])
    ax.set_yticklabels(np.unique(y_labels), fontsize=label_fontsize)

    # x_label = ax.get_xticklabels()
    # [x_label_temp.set_fontname('Times New Roman') for x_label_temp in x_label]
    # y_label = ax.get_yticklabels()
    # [y_label_temp.set_fontname('Times New Roman') for y_label_temp in y_label]

    return ax


def _norm_mat(data, axis, method):
    if axis == 1:
        data = data.T
    if method == 'z_score':
        for i in range(data.shape[1]):
            data[:, i] = (data[:, i] - data[:, i].mean()) / data[:, i].std()
    elif method == 'normalization':
        for i in range(data.shape[1]):
            data[:, i] = (data[:, i] - min(data[:, i])) / \
                                     (max(data[:, i]) - min(data[:, i]))
    elif method == 'proportion':
        for i in range(data.shape[1]):
            data[:, i] = (data[:, i]) / sum(data[:, i])
    if axis == 1:
        data = data.T
    return data


def _consensus_list(mat, ax, metric='euclidean', method='average', axis=0):
    """
    Generate dendrogram.

    Parameters
    ----------
    mat : np.ndarray
        Input matrix.
    ax : Axes
        Matplotlib Axes object.
    metric : str, optional
        Distance metric.
    method : str
        Hierarchical clustering method.
    axis : int, optional
        Clustering axis.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Reordered indices and linkage matrix.
    """
    from scipy.spatial import distance
    from scipy.cluster import hierarchy

    if axis != 0 and axis !=1:
        ValueError('The axis must be either 0 or 1')
    if axis == 1:
        mat = mat.T
    pairwise_dists = distance.pdist(mat, metric=metric)
    linkage = hierarchy.linkage(pairwise_dists, method=method)
    if axis == 0:
        dendrogram = hierarchy.dendrogram(linkage, color_threshold=-np.inf, ax=ax, orientation='left')
    else:
        dendrogram = hierarchy.dendrogram(linkage, color_threshold=-np.inf, ax=ax)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])
    reordered_ind = dendrogram['leaves']
    return reordered_ind, linkage


def _heatmap_with_dendrogram_and_bar(
        data: np.ndarray,
        data_count: np.ndarray,
        x_label: str,
        y_label: str,
        x_map: dict = None,
        y_map: dict = None,
        x_dendrogram: bool = True,
        y_dendrogram: bool = True,
        x_bar: bool = True,
        y_bar: bool = True,
        method: Optional[Literal['z_score', 'normalization', 'proportion']] = 'normalization',
        norm_axis: Optional[Literal[0, 1]] = None,
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        cmap: LinearSegmentedColormap = None,
        xbar_kwags: Mapping[str, Any] = MappingProxyType({}),
        ybar_kwags: Mapping[str, Any] = MappingProxyType({}),
        **kwargs
) -> Tuple[Figure, dict]:
    """
    Display heatmap with two counting barplots.

    Parameters
    ----------
    data : np.ndarray
        Data.
    data_count : np.ndarray
        Count data.
    x_label : str
        x-axis label.
    y_label : str
        y-axis label.
    x_map : Optional[dict], optional
        x-axis mapping.
    y_map : Optional[dict], optional
        y-axis mapping.
    x_dendrogram : bool, optional
        Whether to show x-axis dendrogram.
    y_dendrogram : bool, optional
        Whether to show y-axis dendrogram.
    x_bar : bool, optional
        Whether to show x-axis count barplot.
    y_bar : bool, optional
        Whether to show y-axis count barplot.
    method : Optional[Literal['z_score', 'normalization', 'proportion']], optional
        Normalization method.
    norm_axis : Optional[Literal[0, 1]], optional
        Normalization axis.
    title : Optional[str], optional
        Figure title.
    figsize : Optional[Tuple[float, float]], optional
        Figure size.
    dpi : int, optional
        Image resolution.
    cmap : Optional[LinearSegmentedColormap], optional
        Color map for continuous variables.
    xbar_kwags : Mapping[str, Any], optional
        Parameters for x-axis count barplot.
    ybar_kwags : Mapping[str, Any], optional
        Parameters for y-axis count barplot.
    kwargs
        Other parameters.

    Returns
    -------
    Tuple[Figure, List[Axes]]
        Generated Figure object and list of Axes.
    """
    import seaborn as sns

    norm_data = copy.deepcopy(data)
    axes = {}

    if norm_axis is not None:
        norm_data = _norm_mat(norm_data, norm_axis, method)

    if figsize is None:
        figsize = (12, 8)

    fig = plt.figure(figsize=figsize, dpi=dpi)
    fig.suptitle(title, fontsize=20, x=0.5, y=0.93)

    gs = GridSpec(40, 40)

    ax_heatmap = fig.add_subplot(gs[4:29, 4:29])
    ax_cbar = fig.add_subplot(gs[0:3, 2:3])
    axes['ax_heatmap'] = ax_heatmap
    axes['ax_cbar'] = ax_cbar

    if x_dendrogram:
        ax_dendrogram_x = fig.add_subplot(gs[1:4, 4:29])
        new_xtick, linkage_x = _consensus_list(norm_data, ax_dendrogram_x,
                                               metric='euclidean', method='average', axis=1)
        axes['ax_dendrogram_x'] = ax_dendrogram_x
    else:
        new_xtick = range(norm_data.shape[1])
    if y_dendrogram:
        ax_dendrogram_y = fig.add_subplot(gs[4:29, 1:4])
        new_ytick, linkage_y = _consensus_list(norm_data, ax_dendrogram_y,
                                               metric='euclidean', method='average', axis=0)
        axes['ax_dendrogram_y'] = ax_dendrogram_y
    else:
        new_ytick = range(norm_data.shape[0])

    sns.heatmap(norm_data[np.ix_(new_ytick, new_xtick)],
                ax=ax_heatmap, cmap=cmap, cbar_ax=ax_cbar,
                **kwargs)

    len_x = len(new_xtick)
    len_y = len(new_ytick)
    if x_map is not None:
        dic_new = dict(zip(x_map.values(), x_map.keys()))
        new_xtick = [dic_new[i] for i in new_xtick]
    if y_map is not None:
        dic_new = dict(zip(y_map.values(), y_map.keys()))
        new_ytick = [dic_new[i] for i in new_ytick]

    if x_bar:
        ax_heatmap.set_xticks([])
        xbar_kwags = dict(xbar_kwags)
        bottom_hist = fig.add_subplot(gs[30:38, 4:29])
        sns.countplot(x=x_label, data=data_count, ax=bottom_hist, order=new_xtick, **xbar_kwags)
        # bottom_hist.set_xticks(np.arange(len_x), new_xtick)
        bottom_hist.xaxis.set_ticks_position('top')
        bottom_hist.invert_yaxis()
        bottom_hist.spines['right'].set_visible(False)
        bottom_hist.spines['bottom'].set_visible(False)
        bottom_hist.spines['left'].set_visible(False)
        bottom_hist.set_xlabel(x_label)
    else:
        ax_heatmap.set_xticks(np.arange(len_x), new_xtick)

    if y_bar:
        ax_heatmap.set_yticks([])
        ybar_kwags = dict(ybar_kwags)
        right_hist = fig.add_subplot(gs[4:29, 30:38])
        sns.countplot(y=y_label, data=data_count, ax=right_hist, orient='h', order=new_ytick, **ybar_kwags)
        # right_hist.set_yticks(np.arange(len_y), new_ytick)
        right_hist.set_ylabel(y_label, rotation=270)
        right_hist.yaxis.set_label_position("right")
        right_hist.spines['top'].set_visible(False)
        right_hist.spines['right'].set_visible(False)
        right_hist.spines['bottom'].set_visible(False)
    else:
        ax_heatmap.set_yticks(np.arange(len_y), new_ytick)

    return fig, axes


import os
import numpy as np
import pandas as pd
from typing import Optional, Union
from anndata import AnnData
import numba as nb
import logging as logg


def _assert_variable_in_list(
        variable: str,
        target_list,
        list_name: str,
) -> None:
    if variable not in target_list:
        raise KeyError(f"Spatial basis `{variable}` not found in `{list_name}`.")


def set_R_environment(
        R_HOME: Optional[str] = None,
        R_USER: Optional[str] = None
) -> None:
    """
    Setting R's environment

    Parameters
    ----------
    R_HOME
        PathLike, the location of R (if 'cluster' is m_cluster).
    R_USER
        PathLike, the location of R (if 'cluster' is m_cluster).
    """

    if R_HOME is not None:
        os.environ['R_HOME'] = R_HOME
    if R_USER is not None:
        os.environ['R_USER'] = R_USER


def _best_k(data, k, sdbw):
    """
    The optimal number of clusters is selected by sdbw index
    """
    from sklearn.cluster import KMeans
    from s_dbw import S_Dbw

    data = data.astype(np.float32)

    if sdbw and k != 2:
        s_Dbw = {}
        for k_i in range(2, k+1):
            if data.shape[0] <= k_i:
                break
            km = KMeans(n_clusters=k_i)
            clusters_res = km.fit_predict(data)
            s_Dbw[k_i] = S_Dbw(data, clusters_res)

        min_k = list(s_Dbw.keys())[list(s_Dbw.values()).index(min(list(s_Dbw.values())))]
    else:
        min_k = k

    km = KMeans(n_clusters=min_k)
    res = km.fit_predict(data)
    logg.info(f'The k in cluster is {min_k}', exc_info=True)
    return res, min_k


@nb.jit()
def _filter_of_graph(adata: AnnData,
                     cluster_label: str,
                     exclude: Union[str, dict]) -> AnnData:

    if type(exclude) == str:
        assert exclude in ['same', 'different']
    if type(exclude) == dict:
        exclude = [(i, j) for i in exclude.keys()
                   for j in exclude.values()] + \
                  [(i, j) for j in exclude.keys()
                   for i in exclude.values()]

    distances, neighborhoods = adata.uns['SOAPy']['distance'], adata.uns['SOAPy']['indices']

    for i, neigh in enumerate(neighborhoods):
        point_1 = neigh[0]
        for j, point_2 in enumerate(neigh):
            index_point1 = adata.obs.index[point_1]
            index_point2 = adata.obs.index[point_2]
            if exclude == 'same' and \
                    adata.obs.loc[index_point1, cluster_label] == adata.obs.loc[index_point2, cluster_label]:
                distances[i, j] = -1
                neighborhoods[i, j] = -1
            if exclude == 'different' and \
                    adata.obs.loc[index_point1, cluster_label] != adata.obs.loc[index_point2, cluster_label]:
                distances[i, j] = -1
                neighborhoods[i, j] = -1
            if (adata.obs.loc[index_point1, cluster_label], adata.obs.loc[index_point2, cluster_label]) in exclude:
                distances[i, j] = -1
                neighborhoods[i, j] = -1

    adata.uns['SOAPy']['distance'], adata.uns['SOAPy']['indices'] = distances, neighborhoods
    return adata


# @nb.jit
def _preprocessing_of_graph(adata: AnnData,
                            cluster_label: str,
                            ) -> pd.DataFrame:
    """

    Parameters
    ----------
    adata
    cluster_label

    Returns
    -------

    """

    distances, neighborhoods = adata.uns['SOAPy']['distance'], adata.uns['SOAPy']['indices']
    edges = []

    obs = adata.obs[cluster_label]
    obs_value = obs.values
    for neigh in neighborhoods:
        if len(neigh) == 0:
            continue
        point_1 = neigh[0]
        for point_2 in neigh:
            if point_2 == point_1 | point_2 == -1:
                continue
            elif point_2 < point_1:
                edge = [point_2, point_1,
                        obs_value[point_2],
                        obs_value[point_1]]
                edges.append(edge)

            elif point_2 > point_1:
                edge = [point_1, point_2,
                        obs_value[point_1],
                        obs_value[point_2]]
                edges.append(edge)

    df_edge = pd.DataFrame(data=np.array(edges), columns=['point_1', 'point_2',
                                                          'cluster_1', 'cluster_2'])

    df_edge.drop_duplicates(subset=['point_1', 'point_2'], inplace=True)
    return df_edge


def _count_edge(edge: pd.DataFrame,
                species_of_clusters: int,
                cell_type_dict: dict,
                ) -> np.ndarray:
    """

    Parameters
    ----------
    edge
    species_of_clusters
    cell_type_dict

    Returns
    -------

    """
    enrich_metrics = np.zeros((species_of_clusters, species_of_clusters), dtype=np.float32)

    for row in edge.itertuples():
        enrich_metrics[cell_type_dict[getattr(row, 'cluster_1')], cell_type_dict[getattr(row, 'cluster_2')]] += 1
        enrich_metrics[cell_type_dict[getattr(row, 'cluster_2')], cell_type_dict[getattr(row, 'cluster_1')]] += 1

    return enrich_metrics


def _randomize_helper(adata,
                      cluster_label,
                      species_of_clusters: int,
                      cell_type_dict: dict,
                      ):
    """

    Parameters
    ----------
    adata
    cluster_label
    species_of_clusters
    cell_type_dict

    Returns
    -------

    """
    Series_cluster = adata.obs[cluster_label]
    iter_cluster = Series_cluster.sample(frac=1.0,
                                         ).reset_index(drop=True)
    adata.obs[cluster_label] = iter_cluster.tolist()
    iter_edge = _preprocessing_of_graph(adata,
                                        cluster_label=cluster_label
                                        )

    enrichment = _count_edge(iter_edge, species_of_clusters, cell_type_dict)
    return enrichment


class Iterators:
    def __init__(self,
                 n_iter: int,
                 *args
                 ):
        self.stop = n_iter
        self.params = []
        for param in args:
            self.params.append(param)

    def __iter__(self):
        self.k = 0
        return self

    def __next__(self):
        if self.k < self.stop:
            self.k += 1
            return self.params
        else:
            raise StopIteration


def insert_to_uns(adata, data, *args):
    pointer = adata.uns
    last_index = args[-1]
    args = args[:-1]
    for index in args:
        try:
            pointer[index]
        except KeyError:
            pointer[index] = {}
        pointer = pointer[index]
    pointer[last_index] = data


def adj_pvals(
    data,
    pvalues_data,
    pval_adj_cutoff: float = 0.05,
    correct_axis: str = 'spot',
    adj_method: str = "fdr_bh",
):
    """
    Performs p-value adjustment and determination of significant spots.
    """
    from statsmodels.stats.multitest import multipletests

    scores = data
    sig_scores = scores.copy()
    ps = pvalues_data
    padjs = np.ones(ps.shape)
    if correct_axis == 'spot':
        for spot_i in range(ps.shape[0]):
            lr_indices = np.where(ps[spot_i, :] != 1)[0]
            if len(lr_indices) > 0:
                spot_ps = ps[spot_i, lr_indices]
                spot_padjs = multipletests(spot_ps, method=adj_method)[1]
                padjs[spot_i, lr_indices] = spot_padjs
                sig_scores[spot_i, lr_indices[spot_padjs >= pval_adj_cutoff]] = 0
    elif correct_axis == 'lr':
        for lr_i in range(ps.shape[1]):
            spot_indices = np.where(ps[:, lr_i] != 1)[0]
            if len(spot_indices) > 0:
                lr_ps = ps[spot_indices, lr_i]
                spot_padjs = multipletests(lr_ps, method=adj_method)[1]
                padjs[spot_indices, lr_i] = spot_padjs
                sig_scores[spot_indices[spot_padjs >= pval_adj_cutoff], lr_i] = 0
    else:
        raise Exception(
            f"Invalid correct_axis input, the p value data only have two dimensions"
        )

    # Counting spots significant per lr #
    lr_counts = (padjs < pval_adj_cutoff).sum(axis=0)

    # Re-ranking LRs based on these counts & updating LR ordering #
    new_order = np.argsort(-lr_counts)
    print(f"Updated adata.uns[lr_summary]")
    scores_ordered = scores[:, new_order]
    sig_scores_ordered = sig_scores[:, new_order]
    ps_ordered = ps[:, new_order]
    padjs_ordered = padjs[:, new_order]

    update_p_data = {}
    keys = ["lr_scores", "lr_sig_scores", "p_vals", "p_adjs"]
    values = [scores_ordered, sig_scores_ordered, ps_ordered, padjs_ordered]
    for i in range(len(keys)):
        update_p_data[keys[i]] = values[i]

    return update_p_data


def allocation_edge_2_diff_cell_type(edges, k_max=10000, error=0.001):
    length = len(edges)
    matrix = np.outer(edges, edges) / np.sum(edges)

    for k in range(k_max):
        matrix_new = np.zeros(shape=(length, length), dtype=np.float32)
        sum_ = matrix.trace()

        for i in range(length):
            for j in range(length):
                if i == j:
                    matrix_new[i, j] = matrix[i, j] * matrix[i, j] / sum_
                else:
                    matrix_new[i, j] = matrix[i, j] + (matrix[i, i] * matrix[j, j]) / sum_

        if np.linalg.norm(matrix_new.diagonal()) <= error:
            matrix = matrix_new
            break
        matrix = matrix_new

    if np.linalg.norm(matrix.diagonal()) > error:
        residue_edge = np.max(matrix.diagonal())
        residue_index = np.argmax(matrix.diagonal())
        for i in range(length):
            if i == residue_index:
                matrix[residue_index, residue_index] = 0
            else:
                matrix[residue_index, i] = matrix[i, residue_index] = \
                    residue_edge * matrix[i, residue_index] / (edges[residue_index] - residue_edge)

    return matrix


#!/usr/bin/env python
# -*- coding:utf-8 -*-
from . import pp
from . import tl
from . import pl
from .tl import other_package_without_pip

__version__ = "v0.1.7"

"""
# Author: Lihong lab
# File Name:SOAPy
# github:https://github.com/LiHongCSBLab/SOAPy
# Description:
"""

__author__ = "Cancer system biology lab"
__email__ = [
    "wangheqi2021@sinh.ac.cn"
    "lijiarong2020@sinh.ac.cn"
    "lihong01@sibs.ac.cn"
]






import setuptools
from pathlib import Path

with open("README.md", encoding="utf8") as readme_file:
    readme = readme_file.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name='SOAPy_st',
    version='0.1.7',
    author='Cancer system biology lab',
    author_email='wangheqi2021@sinh.ac.cn',
    description='Spatial Omics Analysis in Python',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/LiHongCSBLab/SOAPy',
    python_requires='>=3.9',
    include_package_data=True,
    package_data={
        "SOAPy_st": ['tl/datasets/Human-2020-Jin-LR-pairs.csv', 'tl/datasets/Mouse-2020-Jin-LR-pairs.csv'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3'],
    keywords=sorted(
        [
            "bio-informatics",
            "spatial omics",
            "spatial transcriptomics",
            "spatial data analysis",
        ]),
    packages=setuptools.find_packages(include=["SOAPy_st", "SOAPy_st.*"]),
    install_requires=requirements
)


import pandas as pd
import numpy as np
import tensorly as tl
import anndata as ad
from anndata import AnnData
from typing import Optional, Union, Literal, Tuple
from ..utils import _check_adata_type
# from utils import insert_to_uns

__all__ = ['TensorDecomposition']


class TensorDecomposition:
    """
    Function:

    The data of anndata type is assembled tensor and tensor decomposition is carried out

    It supports the construction of high-dimensional tensors in count form using attributes in obs and the construction
    of high-dimensional tensors in expression value form using attributes in obs matching genes

    The constructed tensor can use Z-score, 0-1 regularization or proportional form for data balancing

    Tensor decomposition provides two forms of CP decomposition and tucker decomposition, both of which can choose the
    non-negative form of decomposition to assemble tensors of anndata type and conduct tensor decomposition
    """
    def __init__(self):
        """
        Initialize the tensorDecomposition object.
        """

        self.tensor = None
        self.dict_sup = None
        self.factor_name = None
        self.CP = None
        self.tucker = None

    def input_tensor(
            self,
            tensor: np.ndarray,
            factor_name: list,
            ticks_list: list
    ) -> None:
        """
        Set the input tensor, factor names, and supporting dictionary.

        Parameters
        ----------
        tensor : np.ndarray
            The input tensor for decomposition.
        factor_name : list
            A list of factor names.
        ticks_list : list
            A list of tick values corresponding to factor names.

        Returns
        -------
        None
        """
        self.tensor = tensor
        self.factor_name = factor_name
        self.dict_sup = {}
        for index, ticks in enumerate(ticks_list):
            dict_sub = {clu: i for i, clu in enumerate(ticks)}
            self.dict_sup[factor_name[index]] = dict_sub

    @property
    def get_tensor(self) -> np.ndarray:
        """
        Get the current tensor.

        Returns
        -------
        np.ndarray
            The current tensor.
        """
        return self.tensor

    def tensor_with_obs(
            self,
            adata: AnnData,
            obs_factor: list,
    ) -> None:
        """
        Construct a tensor using observation factors from an AnnData object.

        Parameters
        ----------
        adata : AnnData
            An AnnData object containing spatial omics data and spatial information.
        obs_factor : list
            A list of observation factors to use for tensor construction.

        Returns
        -------
        None
        """

        def add_array_value(arr, value, *args):
            arr[args] += value
            return arr

        self.factor_name = obs_factor
        self.dict_sup = {}
        dimensionality = []
        for label in obs_factor:
            dict_sub = {clu: i for i, clu in enumerate(adata.obs[label].unique())}
            self.dict_sup[label] = dict_sub
            dimensionality.append(len(dict_sub))
        tensor = np.zeros(dimensionality)

        for row in adata.obs.itertuples():
            index = []
            for label in obs_factor:
                index.append(self.dict_sup[label][getattr(row, label)])
            add_array_value(tensor, 1, *index)
        self.tensor = tl.tensor(tensor)

    def tensor_with_gene(
            self,
            adata: AnnData,
            obs_factor: list,
            gene_name: Optional[list] = None,
            method: Literal['mean', 'max', 'sum'] = 'mean',
    ) -> None:
        """
        Construct a tensor using observation factors and gene expression values from an AnnData object.

        Parameters
        ----------
        adata : AnnData
            An AnnData object containing spatial omics data and spatial information.
        obs_factor : list
            A list of observation factors to use for tensor construction.
        gene_name : list or None, optional
            A list of gene names to include in the tensor. If None, all genes in adata will be used.
        method : Literal['mean', 'max', 'sum']
            The method of combining expression levels,
            'mean': the mean expression levels of all cells with consistent obs_factor;
            'max': the maximum value of obs_factor expression of all cells consistent with obs_factor;
            'sum': the expression of all cells with consistent obs_factor was summed.

        Returns
        -------
        None
        """
        self.dict_sup = {}
        if gene_name is None:
            gene_name = list(adata.var_names)
        gene_dict = {gene: index for index, gene in enumerate(adata.var_names)}
        dimensionality = []

        # Create mappings for obs_factor and gene
        for label in obs_factor:
            dict_sub = {clu: i for i, clu in enumerate(adata.obs[label].unique())}
            self.dict_sup[label] = dict_sub
            dimensionality.append(len(dict_sub))
        dict_gene = {gene: i for i, gene in enumerate(gene_name)}
        self.dict_sup['gene'] = dict_gene
        dimensionality.append(len(dict_gene))

        # Initialize tensors
        tensor = np.zeros(dimensionality, dtype=np.float32)
        counts = np.zeros(dimensionality, dtype=np.int32)

        # Vectorized computation
        obs_indices = [
            adata.obs[label].map(self.dict_sup[label]).values for label in obs_factor
        ]
        obs_indices = np.array(obs_indices).T  # Shape: (n_cells, len(obs_factor))
        gene_indices = np.array([dict_gene[gene] for gene in gene_name])

        # Prepare index arrays
        obs_indices_expanded = np.repeat(obs_indices, len(gene_name), axis=0)
        gene_indices_expanded = np.tile(gene_indices, obs_indices.shape[0])

        # Flattened index computation for fast access
        flattened_indices = np.ravel_multi_index(
            (
                *obs_indices_expanded.T,
                gene_indices_expanded,
            ),
            tensor.shape,
        )

        # Update tensor and counts
        expression_values = adata.X[:, [gene_dict[gene] for gene in gene_name]].flatten()
        np.add.at(tensor, np.unravel_index(flattened_indices, tensor.shape), expression_values)
        np.add.at(counts, np.unravel_index(flattened_indices, counts.shape), 1)

        # Adjust for 'mean' method
        if method == 'mean':
            counts[counts == 0] = 1  # Prevent division by zero
            tensor /= counts

        self.factor_name = obs_factor + ['gene']
        self.tensor = tl.tensor(tensor)

    def normalization(
            self,
            factor_name: str,
            method: Literal['proportion', 'normalization', 'proportion', 'max'] = 'proportion'
    ) -> None:
        """
        Normalize the tensor along a specified factor.

        Parameters
        ----------
        factor_name : str
            The name of the factor along which normalization should be applied.
        method : Literal['z_score', 'normalization', 'proportion']
            The normalization method.

        Returns
        -------
        None
        """
        tensor = self.tensor
        shape = [index for index, name in enumerate(self.factor_name)]
        axis = self.factor_name.index(factor_name)
        shape[0], shape[axis] = shape[axis], shape[0]
        tensor = np.transpose(tensor, shape)
        dict_index = self.dict_sup[factor_name]
        if method == 'z_score':
            for key, val in dict_index.items():
                axis_mu = np.mean(tensor[val, :, :])
                axis_std = np.std(tensor[val, :, :])
                tensor[val, :, :] = (tensor[val, :, :] - axis_mu) / axis_std
        elif method == 'normalization':
            for key, val in dict_index.items():
                axis_max = np.max(tensor[val, :, :])
                axis_min = np.min(tensor[val, :, :])
                tensor[val, :, :] = (tensor[val, :, :] - axis_min) / (axis_max - axis_min)
        elif method == 'proportion':
            for key, val in dict_index.items():
                cell_number = np.sum(tensor[val, :, :])
                tensor[val, :, :] /= cell_number
        elif method == 'max':
            for key, val in dict_index.items():
                axis_max = np.max(tensor[val, :, :])
                tensor[val, :, :] = tensor[val, :, :] / axis_max
        tensor = np.transpose(tensor, shape)
        self.tensor = tensor

    def highly_variable(
            self,
            factor_name: str,
            top_num: int,
    ):
        tensor = self.tensor
        shape = [index for index, name in enumerate(self.factor_name)]
        axis = self.factor_name.index(factor_name)
        shape[0], shape[axis] = shape[axis], shape[0]
        tensor = np.transpose(tensor, shape)
        dict_index = self.dict_sup[factor_name]

        std = []
        for key, val in dict_index.items():
            std.append(np.std(tensor[val, :, :]))

        rank_std = np.argsort(-np.array(std))[0: top_num]
        tensor = tensor[rank_std, :, :]
        key_ = list(dict_index.keys())
        key_ = [key_[index] for index in rank_std]

        dict_index = {clu: i for i, clu in enumerate(key_)}
        self.dict_sup[factor_name] = dict_index

        tensor = np.transpose(tensor, shape)
        self.tensor = tensor


    def CP_decomposition(
            self,
            rank: int,
            non_negative: bool = False,
            use_hals: bool = False,
            **kwargs
    ):
        """
        Perform CP decomposition on the tensor.

        Parameters
        ----------
        rank : int
            The rank of the decomposition.
        non_negative : bool, optional
            Whether to use non-negative decomposition.
        **kwargs : Any
            additional keyword arguments for the decomposition algorithm.

        Returns
        -------
        tuple
            A tuple containing weights, factors, and normalized root mean squared error (NRE).
        """
        tensor = self.tensor
        self.CP = {}
        if non_negative:
            if use_hals:
                from tensorly.decomposition import non_negative_parafac_hals
                weights, factors = non_negative_parafac_hals(tensor=tensor,
                                                             rank=rank,
                                                             normalize_factors=True,
                                                             **kwargs)
            else:
                from tensorly.decomposition import non_negative_parafac
                weights, factors = non_negative_parafac(tensor=tensor,
                                                        rank=rank,
                                                        normalize_factors=True,
                                                        **kwargs)
        else:
            from tensorly.decomposition import parafac_power_iteration
            weights, factors = parafac_power_iteration(tensor=tensor,
                                                       rank=rank,
                                                       **kwargs)

        from tensorly.cp_tensor import cp_to_tensor
        # print(weights)
        # print(factors)
        tensor_hat = cp_to_tensor((weights, factors))
        nre = _nre_similar(tensor, tensor_hat)
        self.CP['rank'] = rank
        self.CP['weights'] = weights
        self.CP['factors'] = {}
        for index, name in enumerate(self.factor_name):
            self.CP['factors'][name] = factors[index]
        self.CP['nre'] = nre

        return weights, factors, nre

    def tucker_decomposition(
            self,
            rank: Union[int, tuple, list],
            non_negative: bool = False,
            **kwargs
    ):
        """
        Perform Tucker decomposition on the tensor.

        Parameters
        ----------
        rank : int, tuple, list
            The rank of the decomposition for each mode.
        non_negative : bool, optional
            Whether to use non-negative decomposition.
        **kwargs :
            additional keyword arguments for the tensorly.decomposition algorithm.

        Returns
        -------
        tuple
            A tuple containing core tensor, factors, and normalized root mean squared error (NRE).
        """
        tensor = self.tensor
        self.tucker = {}
        if non_negative:
            from tensorly.decomposition import non_negative_tucker
            core, factors = non_negative_tucker(tensor=tensor, rank=rank, **kwargs)
        else:
            from tensorly.decomposition import tucker
            core, factors = tucker(tensor=tensor, rank=rank, **kwargs)

        from tensorly.tucker_tensor import tucker_to_tensor
        tensor_hat = tucker_to_tensor((core, factors))
        nre = _nre_similar(tensor, tensor_hat)
        self.tucker['rank'] = {}
        if isinstance(rank, int):
            rank = [rank]*len(self.factor_name)
        for index, name in enumerate(self.factor_name):
            self.tucker['rank'][name] = rank[index]
        self.tucker['weights'] = core
        self.tucker['factors'] = {}
        for index, name in enumerate(self.factor_name):
            self.tucker['factors'][name] = factors[index]
        self.tucker['nre'] = nre

        return core, factors, nre


def _nre_similar(tensor1, tensor2) -> float:
    """
    normalized reconstruction error
    """
    # normalized reconstruction error
    from tensorly import norm
    NRE = norm(tensor1 - tensor2)/norm(tensor1)
    return NRE




import copy
import anndata
import numpy as np
import pandas as pd
import logging as logg
from anndata import AnnData
from typing import Optional, Union, Any


def _add_info_from_sample(
        adata: AnnData,
        sample_id: Optional[str] = None,
        keys: Union[str, list, None] = None,
        add: Any = None,
) -> AnnData:
    """
    Storing information
    """

    if isinstance(keys, str):
        keys = [keys]
        add = [add]

    try:
        adata.uns['SOAPy']
    except KeyError:
        adata.uns['SOAPy'] = {}
        logg.warning('adata has not been initialized, and adata.uns[\'SOAPy\'] has been established')

    if sample_id is None:
        for index, key in enumerate(keys):
            adata.uns['SOAPy'][key] = add[index]
            logg.info(f'{key} information has been incorporated into adata.uns[\'SOAPy\'][{key}]')
    else:
        try:
            adata.uns['SOAPy'][sample_id]
        except KeyError:
            adata.uns['SOAPy'][sample_id] = {}
        for index, key in enumerate(keys):
            adata.uns['SOAPy'][sample_id][key] = add[index]
            logg.info(f'{key} information from {sample_id} has been incorporated into adata.uns[\'SOAPy\'][{sample_id}][{key}]')

    return adata


def _get_info_from_sample(
        adata: AnnData,
        sample_id: Optional[str] = None,
        key: Optional[str] = None,
        printf: bool = True
):
    """
    Getting information
    """

    if 'SOAPy' not in adata.uns.keys():
        if printf:
            logg.error('SOAPy was not found in adata.uns', exc_info=True)
        raise KeyError()

    if sample_id is None:
        if key not in adata.uns['SOAPy'].keys():
            if printf:
                logg.error(f'{key} analysis of {sample_id} was not present, please conduct {key} analysis first',
                           exc_info=True)
            raise KeyError()
        data = adata.uns['SOAPy'][key]
    else:
        if key not in adata.uns['SOAPy'][sample_id].keys():
            if printf:
                logg.error(f'{key} analysis of {sample_id} was not present, please conduct {key} analysis first',
                           exc_info=True)
            raise KeyError()
        data = adata.uns['SOAPy'][sample_id][key]

    return data


def _scale(
        adata: AnnData,
        library_id: Optional[str] = None,
        res: Optional[str] = None,
) -> float:
    """
    Get the scale of the coordinates to the drawing map
    """
    try:
        if library_id is None:
            library_id = list(adata.uns['spatial'].keys())[0]
        scale = adata.uns['spatial'][library_id]['scalefactors']['tissue_' + res + '_scalef']
    except KeyError:
        scale = 1.0

    return scale


def _neighbor_network(
        points: np.ndarray,
        max_quantile: int = 98,
):

    def map_dis_to_points(point_pairs, dis_matrix, points_num):
        indices = [[] for i in range(points_num)]
        distances = [[] for i in range(points_num)]
        for pair in point_pairs:
            pi, pj, pk = pair
            indices[pi].append(pj)
            indices[pi].append(pk)
            indices[pj].append(pi)
            indices[pj].append(pk)
            indices[pk].append(pi)
            indices[pk].append(pj)
        indices = [list(set(neigh)) for neigh in indices]
        for index, neighs in enumerate(indices):
            for point in neighs:
                distances[index].append(dis_matrix[index, point])
        return distances, indices

    def filter_upper(distances, indices, max_):
        dis_flatten = []
        for dis_neigh in distances:
            dis_flatten += dis_neigh
        max = np.percentile(dis_flatten, max_)
        for pi, neigh in enumerate(indices):
            for index, pj in enumerate(neigh):
                if distances[pi][index] > max:
                    distances[pi][index] = -1
                    indices[pi][index] = -1
        distances = [[dis for dis in dis_n if not dis < 0] for dis_n in distances]
        indices = [[pointj for pointj in nei_n if not pointj < 0] for nei_n in indices]
        return distances, indices

    from scipy.spatial import Delaunay
    from scipy.spatial import distance
    import numpy as np

    tri = Delaunay(points)

    points = np.array(points, dtype=np.float32)
    dis_matrix = distance.cdist(points, points, 'euclidean')

    distances, indices = map_dis_to_points(tri.simplices, dis_matrix, len(points))
    distances, indices = filter_upper(distances, indices, max_quantile)

    distances = np.array([np.array(dis) for dis in distances], dtype=object)
    indices = np.array([np.array(ind) for ind in indices], dtype=object)
    return distances, indices


def _graph(col, row, method, cutoff=None, max_quantile=None):
    """
    Build an adjacent network

    """
    from sklearn.neighbors import NearestNeighbors

    def _KNN_neighbors(crd, k):
        nbrs = NearestNeighbors(n_neighbors=k + 1, metric="euclidean")
        nbrs.fit(crd)
        distances, indices = nbrs.kneighbors(crd)
        return distances, indices

    def _radius_neighbors(crd, radius):
        nbrs = NearestNeighbors(radius=radius, metric="euclidean")
        nbrs.fit(crd)
        distances, indices = nbrs.radius_neighbors(crd)
        return distances, indices

    X = list(zip(col, row))
    if method == 'knn':
        distances, indices = _KNN_neighbors(crd=X, k=cutoff)
    elif method == 'radius':
        distances, indices = _radius_neighbors(crd=X, radius=cutoff)
    elif method == 'regular':
        distances, indices = _KNN_neighbors(crd=X, k=cutoff)
        dist = distances.flatten()
        dist_cutoff = np.median(dist) * 1.2
        mask = np.where(distances < dist_cutoff, True, False)

        distances_correct = []
        indices_correct = []
        for row in range(mask.shape[0]):
            auxiliary_dist = []
            auxiliary_indices = []
            for col in range(mask.shape[1]):
                if mask[row, col]:
                    auxiliary_dist.append(distances[row, col])
                    auxiliary_indices.append(indices[row, col])
            distances_correct.append(np.array(auxiliary_dist))
            indices_correct.append(np.array(auxiliary_indices))
        distances = distances_correct
        indices = indices_correct
    elif method == 'neighbor':
        distances, indices = _neighbor_network(np.array(list(zip(col, row))), max_quantile)
    else:
        raise ValueError("The method must be in ['radius', 'knn', 'regular' ,'neighbor']")

    return indices, distances


def _check_adata_type(
        adata: anndata.AnnData,
        spatial_in_obsm: str,
        inplace: bool,
) -> anndata.AnnData:

    from scipy.sparse import csr_matrix

    if not inplace:
        adata = copy.deepcopy(adata)

    adata.X = csr_matrix(adata.X)

    if spatial_in_obsm in adata.obsm.keys():
        if isinstance(adata.obsm[spatial_in_obsm], pd.DataFrame):
            adata.obsm[spatial_in_obsm] = adata.obsm[spatial_in_obsm].values
    else:
        pass

    return adata


import logging
import anndata
import numpy as np
import pandas as pd
from typing import Union, Literal
from ..utils import _check_adata_type

__all__ = ["cal_spatialDE", "cal_sparkX", "cal_spark"]


def _get_gene_list(
        gene_name,
        all_genes,
) -> list:

    if gene_name is None:
        gene_name = all_genes
    if isinstance(gene_name, str):
        gene_name = [gene_name]

    is_in_list = np.isin(gene_name, all_genes)
    if not all(is_in_list):
        gene_not_in_list = gene_name[is_in_list]
        logging.error(f'{gene_not_in_list} not in adata.var_names.', exc_info=True)
    return gene_name


def cal_spatialDE(
        adata: anndata.AnnData,
        gene_name: Union[list, str, None] = None,
        spatial_in_obsm: str = 'spatial',
) -> pd.DataFrame:
    """
    Calculate SpatialDE analysis on spatial omics data.
    Doi: 10.1038/nmeth.4636
    Github: https://github.com/Teichlab/SpatialDE

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    gene_name : Union[list, str, None], optional
        List of genes of interest or a single gene name to focus the analysis on.
        If None, the analysis will consider all genes in the dataset.
    spatial_in_obsm : str
        The key of spatial coordinates in adata.obsm

    Returns
    -------
    pd.DataFrame
        A Pandas DataFrame containing the results of the SpatialDE analysis.

    """
    import SpatialDE

    adata = _check_adata_type(adata, spatial_in_obsm, True)

    gene_name = _get_gene_list(gene_name=gene_name, all_genes=adata.var_names.tolist())

    exp_mat = pd.DataFrame(adata.X.todense(), columns=gene_name, index=adata.obs_names)
    coord = adata.obsm[spatial_in_obsm]
    results = SpatialDE.run(coord, exp_mat)

    return results


def cal_sparkX(
        adata: anndata.AnnData,
        gene_name: Union[list, str, None] = None,
        num_core: int = 5,
        spatial_in_obsm: str = 'spatial',
) -> pd.DataFrame:
    """
    Calculate SPARK-X analysis on spatial omics data.
    Doi: 10.1186/s13059-021-02404-0
    Github: https://github.com/xzhoulab/SPARK

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    gene_name : Union[list, str, None], optional
        List of genes of interest or a single gene name to focus the analysis on.
        If None, the analysis will consider all genes in the dataset.
    num_core : int
        The number of CPU core.
    spatial_in_obsm : str
        The key of spatial coordinates in adata.obsm

    Returns
    -------
    pd.DataFrame
        A Pandas DataFrame containing the results of the SPARK-X analysis.

    """

    # load required packages
    from rpy2.robjects import r
    from rpy2.robjects import pandas2ri
    import rpy2.robjects as ro
    pandas2ri.activate()

    adata = _check_adata_type(adata, spatial_in_obsm, True)

    gene_name = _get_gene_list(gene_name=gene_name, all_genes=adata.var_names.tolist())
    count_mat = pd.DataFrame(adata.X.todense(), columns=gene_name, index=adata.obs_names).transpose()
    coord_mat = adata.obsm[spatial_in_obsm]

    RunSPARKX = r(f"""
        function(count_mat,coord_mat){{
        library(SPARK)
        count_mat=as.matrix(count_mat)
        count_mat=Matrix::Matrix(count_mat,sparse=TRUE)
        sparkX_result=sparkx(count_mat,as.matrix(coord_mat),numCores={num_core},option="mixture")
        return(list(data.frame(sparkX_result[["res_mtest"]],check.names=FALSE),rownames(count_mat)))
    }}
    """)

    test_result = RunSPARKX(count_mat, coord_mat)
    test_result[0].index = test_result[1]
    test_result = test_result[0]
    test_result = ro.conversion.rpy2py(test_result)

    return test_result


def cal_spark(
        adata: anndata.AnnData,
        gene_name: Union[list, str, None] = None,
        method: Literal['gaussian', 'poisson'] = 'gaussian',
        num_core: int = 5,
        spatial_in_obsm: str = 'spatial',
) -> pd.DataFrame:
    """
    Perform SPARK analysis on spatial omics data.
    Doi: 10.1038/s41592-019-0701-7
    Github: https://github.com/xzhoulab/SPARK

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    gene_name : Union[list, str, None], optional
        List of genes of interest or a single gene name to focus the analysis on.
        If None, the analysis will consider all genes in the dataset.
    method : Literal['gaussian', 'poisson'], optional
        The statistical method to be used for the analysis. Must be one of 'gaussian' or 'poisson'.
    num_core : int
        The number of CPU core.
    spatial_in_obsm : str
        The key of spatial coordinates in adata.obsm

    Returns
    -------
    pd.DataFrame
        A Pandas DataFrame containing the results of the SPARK analysis.
    """

    # load required packages
    from rpy2.robjects import r
    from rpy2.robjects import pandas2ri
    import rpy2.robjects as ro
    pandas2ri.activate()

    adata = _check_adata_type(adata, spatial_in_obsm, True)

    gene_name = _get_gene_list(gene_name=gene_name, all_genes=adata.var_names.tolist())

    count_mat = pd.DataFrame(adata.X.todense(), columns=gene_name, index=adata.obs_names).transpose()
    coord_mat = adata.obsm[spatial_in_obsm]

    def generate_spark_r_function(method):
        r_code = f"""
            function(count_mat, coord_mat) {{
                library(SPARK)
                rownames(coord_mat) = colnames(count_mat)
                spark_obj = CreateSPARKObject(counts = count_mat, location = coord_mat, percentage = 0, min_total_counts = 0)
                spark_obj@lib_size = apply(spark_obj@counts, 2, sum)
                spark_obj = spark.vc(spark_obj, covariates = NULL, fit.model = '{method}', num_core = '{num_core}', verbose = F)
                spark_obj = spark.test(spark_obj, check_positive = T, verbose = F)
                return(spark_obj@res_mtest)
            }}
        """
        return r_code

    test_result = r(generate_spark_r_function(method=method))(count_mat, coord_mat)
    test_result = ro.conversion.rpy2py(test_result)

    return test_result